from __future__ import absolute_import, print_function

import requests

from sentry.auth.providers.oauth2 import (
    OAuth2Callback, OAuth2Provider, OAuth2Login
)
from .constants import (
    AUTHORIZATION_ENDPOINT,
    USERINFO_ENDPOINT,
    ISSUER, TOKEN_ENDPOINT,
    CLIENT_SECRET,
    CLIENT_ID,
    SCOPE, DATA_VERSION,
    PROJECT,
    AUTH_URL
)
from .views import FetchUser, OIDCConfigureView
import logging
logger = logging.getLogger('sentry.auth.oidc')


class OIDCLogin(OAuth2Login):
    authorize_url = AUTHORIZATION_ENDPOINT
    client_id = CLIENT_ID
    scope = SCOPE

    def __init__(self, domains=None):
        self.domains = domains
        super(OIDCLogin, self).__init__()

    def get_authorize_params(self, state, redirect_uri):
        params = super(OIDCLogin, self).get_authorize_params(
            state, redirect_uri
        )
        # TODO(dcramer): ideally we could look at the current resulting state
        # when an existing auth happens, and if they're missing a refresh_token
        # we should re-prompt them a second time with ``approval_prompt=force``
        params['approval_prompt'] = 'force'
        params['access_type'] = 'offline'
        return params


class OIDCProvider(OAuth2Provider):
    name = ISSUER
    client_id = CLIENT_ID
    client_secret = CLIENT_SECRET

    def __init__(self, domain=None, domains=None, version=None, **config):
        if domain:
            if domains:
                domains.append(domain)
            else:
                domains = [domain]
        self.domains = domains
        # if a domain is not configured this is part of the setup pipeline
        # this is a bit complex in Sentry's SSO implementation as we don't
        # provide a great way to get initial state for new setup pipelines
        # vs missing state in case of migrations.
        if domains is None:
            version = DATA_VERSION
        else:
            version = None
        self.version = version
        super(OIDCProvider, self).__init__(**config)

    def get_configure_view(self):
        return OIDCConfigureView.as_view()

    def get_auth_pipeline(self):
        return [
            OIDCLogin(domains=self.domains),
            OAuth2Callback(
                access_token_url=TOKEN_ENDPOINT,
                client_id=self.client_id,
                client_secret=self.client_secret,
            ),
            FetchUser(
                domains=self.domains,
                version=self.version,
            ),
        ]

    def get_refresh_token_url(self):
        return TOKEN_ENDPOINT

    def build_config(self, state):
        return {
            'domains': [state['domain']],
            'version': DATA_VERSION,
        }

    def get_user_info(self, bearer_token):
        endpoint = USERINFO_ENDPOINT
        bearer_auth = 'Bearer ' + bearer_token
        return requests.get(endpoint + "?schema=openid",
                            headers={'Authorization': bearer_auth},
                            timeout=2.0).json()

    def build_identity(self, state):
        data = state['data']
        bearer_token = data['access_token']
        user_info = self.get_user_info(bearer_token)
        if not user_info.get('email'):
            logger.error('Missing email in user endpoint: %s' % data)

        user_data = state['user']
        if self.email_verified(user_info.get("email"), bearer_token):
            return {
                'id': user_data.get('sub'),
                'email': user_info.get('email'),
                'email_verified': user_info.get('email_verified'),
                'nickname': user_info.get('nickname'),
                'name': user_info.get('email'),
                'data': self.get_oauth_data(data),
            }
        return {}

    def email_verified(self, email, bearer_token):
        import time
        r = requests.get(AUTH_URL + "/users?groups.project="+ PROJECT,
            headers={"X-Auth-project":PROJECT,"X-Access-token": bearer_token})
        for retry in range(10):
            if 10 < retry:
                return {}
            if not str(r.status_code).startswith("20"):
                wait_time = 2**retry * 0.1
                time.sleep(wait_time)
                continue
            else:
                for i in r.json()['items']:
                    if i["email"] == email:
                        return True
                return False


from __future__ import absolute_import, print_function

import logging

from sentry.auth.view import AuthView, ConfigureView
from sentry.utils import json

from .constants import (
    ERR_INVALID_RESPONSE,
    ISSUER
)
from sentry.utils.signing import urlsafe_b64decode

logger = logging.getLogger('sentry.auth.oidc')


class FetchUser(AuthView):
    def __init__(self, domains, version, *args, **kwargs):
        self.domains = domains
        self.version = version
        super(FetchUser, self).__init__(*args, **kwargs)

    def dispatch(self, request, helper):
        data = helper.fetch_state('data')

        try:
            id_token = data['id_token']
        except KeyError:
            logger.error('Missing id_token in OAuth response: %s' % data)
            return helper.error(ERR_INVALID_RESPONSE)

        try:
            _, payload, _ = map(urlsafe_b64decode, id_token.split('.', 2))
        except Exception as exc:
            logger.error(u'Unable to decode id_token: %s' % exc, exc_info=True)
            return helper.error(ERR_INVALID_RESPONSE)

        try:
            payload = json.loads(payload)
        except Exception as exc:
            logger.error(u'Unable to decode id_token payload: %s' % exc, exc_info=True)
            return helper.error(ERR_INVALID_RESPONSE)

        if not payload.get('sub'):
            logger.error('Missing sub in id_token payload: %s' % id_token)
            return helper.error(ERR_INVALID_RESPONSE)

        domain = payload.get('iss')

        helper.bind_state('domain', domain)
        helper.bind_state('user', payload)
        return helper.next_step()


class OIDCConfigureView(ConfigureView):
    def dispatch(self, request, organization, auth_provider):
        config = auth_provider.config
        if config.get('domain'):
            domains = [config['domain']]
        else:
            domains = config.get('domains')
        return self.render('oidc/configure.html', {
            'provider_name': ISSUER or "",
            'domains': domains or []
        })


class FetchDetailedUserInfo(AuthView):
    def __init__(self, domains, version, *args, **kwargs):
        self.domains = domains
        self.version = version
        super(FetchDetailedUserInfo, self).__init__(*args, **kwargs)

    def dispatch(self, request, helper):
        data = helper.fetch_state('data')

        try:
            id_token = data['id_token']
        except KeyError:
            logger.error('Missing id_token in OAuth response: %s' % data)
            return helper.error(ERR_INVALID_RESPONSE)

        try:
            _, payload, _ = map(urlsafe_b64decode, id_token.split('.', 2))
        except Exception as exc:
            logger.error(u'Unable to decode id_token: %s' % exc, exc_info=True)
            return helper.error(ERR_INVALID_RESPONSE)

        try:
            payload = json.loads(payload)
        except Exception as exc:
            logger.error(u'Unable to decode id_token payload: %s' % exc, exc_info=True)
            return helper.error(ERR_INVALID_RESPONSE)

        if not payload.get('sub'):
            logger.error('Missing sub in id_token payload: %s' % id_token)
            return helper.error(ERR_INVALID_RESPONSE)

        domain = payload.get('iss')

        helper.bind_state('domain', domain)
        helper.bind_state('user', payload)
        return helper.next_step()



from __future__ import absolute_import
from django.apps import AppConfig


class Config(AppConfig):
    name = "oidc"

    def ready(self):
        from sentry import auth
        from .provider import OIDCProvider

        auth.register('oidc', OIDCProvider)



#!/usr/bin/env python
from __future__ import absolute_import
from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

install_requires = [
    'requests<3.0.0'
]

tests_require = [
    'flake8<3.6.0,>=3.5.0',
]

setup(
    name='da2-sentry-auth-oidc',
    version='3.0.8',
    author='Aixin',
    author_email='aixinmys@163.com',
    url='https://www.getsentry.com',
    description='OpenID Connect authentication provider for Sentry',
    long_description=readme,
    license='Apache 2.0',
    packages=find_packages(exclude=['tests']),
    zip_safe=False,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={'tests': tests_require},
    include_package_data=True,
    entry_points={
        'sentry.apps': [
            'oidc = oidc.apps.Config',
        ],
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'Topic :: Software Development',
    ],
)
